# -*- coding: utf-8 -*-
#
#  textmanager.py - TextBuffer manipulator part for GBottler
#  Copyright (C) 2004 by Atzm WATANABE <sitosito@p.chan.ne.jp>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License (version 2) as
#  published by the Free Software Foundation.  It is distributed in the
#  hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the GNU General Public License for more details.
#
# $Id: textmanager.py,v 1.42 2009/12/28 09:59:51 atzm Exp $
#

import os
import sys
import gobject, gtk, pango

from bottleparser import *
from types import *
from common import *
import config

class TextManager(gtk.TextView):
	__gsignals__ = {
		'enter-pressed': (gobject.SIGNAL_RUN_LAST|gobject.SIGNAL_ACTION, gobject.TYPE_NONE, (gobject.TYPE_INT,))
	}

	def __init__(self, app, parser=BottleParser('loose')):
		gtk.TextView.__init__(self)

		self.app = app
		self.parser = parser
		self.view_style = STYLE_TALK
		self.textbuffer = gtk.TextView.get_buffer(self)

		# settings
		self.set_wrap_mode(gtk.WRAP_CHAR)

		# create tags
		self.textbuffer.create_tag('script')
		self.textbuffer.create_tag('sakura')
		self.textbuffer.create_tag('kero')
		self.textbuffer.create_tag('sync')
		self.textbuffer.create_tag('error')
		self.textbuffer.create_tag('overURL')

		url_tag = self.textbuffer.create_tag('URL')
		self.font_tag = self.textbuffer.create_tag('font')

		# some connection
		self.p_flag = False
		url_tag.connect('event', self.receive_event)
		self.connect('motion-notify-event', self.motion)
		self.app.window.connect('focus-out-event', self.focus_out)

		# initialize
		self.receive_preferences_changed()

	def do_enter_pressed(self, modifier):
		if modifier == gtk.gdk.SHIFT_MASK:
			action = config.get('key', 'shift_enter', 'int')
		elif modifier == gtk.gdk.CONTROL_MASK:
			action = config.get('key', 'ctrl_enter', 'int')
		elif modifier == gtk.gdk.MOD1_MASK:
			action = config.get('key', 'alt_enter', 'int')
		else:
			action = config.get('key', 'enter', 'int')

		if action == config.KEY_ENTER_BACKSLASH_N:
			self.emit('insert-at-cursor', r'\n')
		elif action == config.KEY_ENTER_BACKSLASH_N_NEWLINE:
			self.emit('insert-at-cursor', '\\n\n')
		elif action == config.KEY_ENTER_NEWLINE:
			self.emit('insert-at-cursor', '\n')

	def color_changed(self, over_color='light gray'):
		tag_table = self.textbuffer.get_tag_table()

		tag_table.lookup('script').set_property('foreground', config.get('fgcolor', 'script'))
		tag_table.lookup('sakura').set_property('foreground', config.get('fgcolor', 'sakura'))
		tag_table.lookup('kero').set_property('foreground', config.get('fgcolor', 'unyuu'))
		tag_table.lookup('sync').set_property('foreground', config.get('fgcolor', 'synchronized'))
		tag_table.lookup('error').set_property('foreground', config.get('fgcolor', 'error'))

		url_tag = tag_table.lookup('URL')
		url_tag.set_property('foreground', config.get('fgcolor', 'url'))
		url_tag.set_property('underline', pango.UNDERLINE_SINGLE)

		tag_table.lookup('overURL').set_property('background', over_color)

	def font_changed(self):
		self.set_style(self.view_style)

	def receive_preferences_changed(self):
		self.color_changed()
		self.font_changed()
		self.coloring()

	def receive_event(self, texttag, widget, event, iter):
		buf = iter.get_buffer()
		end = iter.copy()

		if not iter.begins_tag(texttag):
			iter.backward_to_tag_toggle(texttag) 

		if not end.ends_tag(texttag):
			end.forward_to_tag_toggle(texttag)

		if event.type == 4 and \
			   event.button and event.button == 1: # 4 left button pressed
			os.system(config.get('browser', 'browser_command') % buf.get_text(iter, end) + ' &')
			buf.remove_tag_by_name('overURL', iter, end)
		elif event.type == 3: # 3 mouse over?
			self.p_flag = True
			buf.apply_tag_by_name('overURL', iter, end)
			event.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.HAND1))

	def focus_out(self, widget, event):
		self.p_flag = False
		self.textbuffer.remove_tag_by_name('overURL', self.textbuffer.get_start_iter(),
										   self.textbuffer.get_end_iter())
		event.window.set_cursor(None)

	def motion(self, widget, event):
		event.window.get_pointer() # why I must call this method?
		if not self.p_flag:
			self.textbuffer.remove_tag_by_name('overURL', self.textbuffer.get_start_iter(),
											   self.textbuffer.get_end_iter())
			event.window.set_cursor(None)
		else:
			self.p_flag = False

	def get_script(self):
		[s, e] = self.textbuffer.get_bounds()
		return self.textbuffer.get_text(s, e)

	RE_NEWLINE = re.compile('\n+$')
	def coloring(self, widget=None, event=None):
		current = ['sakura'] # tag stack for closable (currently only synchronized) sessions
		[s, e] = self.textbuffer.get_bounds()

		string = unicode(self.textbuffer.get_text(s, e), 'utf-8')
		string = self.RE_NEWLINE.sub('', string)

		pos = self.textbuffer.get_iter_at_mark(self.textbuffer.get_insert()).get_offset() # cursor position
		end = self.textbuffer.get_end_iter().get_offset()

		chunks = self.parser.parse(string)

		self.textbuffer.set_text("")
		for chunk in chunks:
			if chunk[0] == TEXT_STRING:
				for n in xrange(len(chunk[1])):
					if chunk[1][n][0] in [SCRIPT_TEXT, TEXT_META]:
						c = chunk[1][n][1]
						self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
															 c, current[-1])
					elif chunk[1][n][0] == PARSE_ERROR:
						c = chunk[1][n][1]
						self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
															 c, 'error')

			elif chunk[0] in [TEXT_META, PARSE_ERROR]:
				tag = 'script'
				name, args = chunk[1], chunk[2:]

				if name in [r'\0', r'\h']:
					if not current[-1] == 'sync':
						current = ['sakura']
				elif name in [r'\1', r'\u']:
					if not current[-1] == 'sync':
						current = ['kero']
				elif name == r'\_s':
					if current[-1] == 'sync' and len(current) > 1:
						# when synchronized session closed
						current.pop()
					else:
						current.append('sync')

				if chunk[0] == PARSE_ERROR:
					tag = 'error'

				self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
													 name, tag)

				if name == r'\URL' and chunk[0] == PARSE_ERROR and len(args) >= 2:
					format = '[%s]'
				elif name == r'\w':
					format = '%s'
				elif name == r'\s' and chunk[0] == PARSE_ERROR and (not args[0] or args[0][0] == '['):
					format = '%s'
				else:
					format = '[%s]'

				for n in xrange(len(args)):
					if type(args[0]) is TupleType:
						try:
							if args[n][0][0] == PARSE_ERROR:
								self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																	 "%s" % args[n][0][1], 'error')
							else:
								self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																	 format % args[n][0][1], tag)
						except IndexError:
							pass
					else:
						self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
															 format % args[n], tag)

		[si, ei] = self.textbuffer.get_bounds()
		self.textbuffer.apply_tag_by_name('font', si, ei)

		if end != pos:
			if end != 0:
				self.emit('move-cursor', gtk.MOVEMENT_LOGICAL_POSITIONS, end*-1, False)
			if pos != 0:
				self.emit('move-cursor', gtk.MOVEMENT_LOGICAL_POSITIONS, pos, False)

	def insert_with_color_from(self, string, pos=-3):
		current = ['sakura'] # tag stack for closable (currently only synchronized) sessions
		for chunk in self.parser.parse(string):
			if chunk[0] == TEXT_STRING:
				if chunk[1][0][0] in [SCRIPT_TEXT, TEXT_META]:
					c = chunk[1][0][1]
					if self.view_style == STYLE_TALK:
						c = c.replace(r"\\", "\\")
						c = c.replace(r"\%", "%")
						c = c.replace(r"\]", "]")
					self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
															 c, current[-1])

			elif chunk[0] in [TEXT_META, PARSE_ERROR]:
				tag = 'script'
				name, args = chunk[1], chunk[2:]

				if name in [r'\0', r'\h']:
					if not current[-1] == 'sync':
						current = ['sakura']
				elif name in [r'\1', r'\u']:
					if not current[-1] == 'sync':
						current = ['kero']
				elif name == r'\_s':
					if current[-1] == 'sync' and len(current) > 1:
						# when synchronized session closed
						current.pop()
					else:
						current.append('sync')

				if chunk[0] == PARSE_ERROR:
					tag = 'error'

				if self.view_style == STYLE_TALK:
					pre_end = self.textbuffer.get_end_iter()
					first = pre_end.backward_char()
					if name in [r'\0', r'\h', r'\1', r'\u', r'\_s'] and \
						   pre_end.get_char() != '\n' and first:
						if current[-1] != 'sync' or name == r'\_s':
							self.textbuffer.insert(self.textbuffer.get_end_iter(), '\n')
					elif name == r'\URL':
						for n in xrange(len(args)):
							if not n % 2 and len(args) != 1:
								self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																		 "[%s]" % args[n][0][1], tag)
							else:
								self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																		 "[", tag)
								self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																		 "%s" % args[n][0][1], 'URL')
								self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																		 "]", tag)

				else:
					if name in [r'\0',  r'\h', r'\1', r'\u', r'\_s',
								r'\_q', r'\t', r'\c', r'\e'] and \
								len(args) == 0:
						self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																 name, tag)
						continue

					if name == r'\URL':
						self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																 name, tag)
					if name == r'\w':
						format = '%s%s'
					else:
						format = '%s[%s]'

					for n in xrange(len(args)):
						if type(args[0]) is TupleType:
							if not n % 2 and len(args) != 1:
								self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																		 '[%s]' % args[n][0][1], tag)
							else:
								if name == r'\URL':
									self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																			 '[', tag)
									self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																			 '%s' % args[n][0][1], 'URL')
									self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																			 ']', tag)
								else:
									tmp = '%s[%s]' % (name, args[n][0][1])
									self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																			 tmp, tag)

								if self.view_style == STYLE_SCRIPT_WITH_LINEFEED and name == r'\n':
									self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																			 '\n', tag)
						else:
							self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(),
																	 format % (name, args[n]), tag)

					if name == r'\n' and not args:
						self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(), r'\n', tag)
						if self.view_style == STYLE_SCRIPT_WITH_LINEFEED:
							self.textbuffer.insert_with_tags_by_name(self.textbuffer.get_end_iter(), '\n', tag)

		self.textbuffer.insert(self.textbuffer.get_end_iter(), '\n')
		[si, ei] = self.textbuffer.get_bounds()
		self.textbuffer.apply_tag_by_name('font', si, ei)

		if pos != 0:
			self.emit("move-cursor", gtk.MOVEMENT_LOGICAL_POSITIONS, pos, False)

	def set_style_talk(self):
		self.view_style = STYLE_TALK
		self.font_tag.set_property('font', config.get('font', 'general'))

	def set_style_script(self):
		self.view_style = STYLE_SCRIPT
		self.font_tag.set_property('font', config.get('font', 'script'))

	def set_style_script_with_linefeed(self):
		self.view_style = STYLE_SCRIPT_WITH_LINEFEED
		self.font_tag.set_property('font', config.get('font', 'script'))

	def set_style(self, style):
		if type(style) is not IntType:
			raise ValueError("Invalid argument: " + str(style))

		if style == STYLE_TALK:
			self.set_style_talk()
		elif style == STYLE_SCRIPT:
			self.set_style_script()
		else:
			self.set_style_script_with_linefeed()

	def clear(self):
		self.textbuffer.set_text("")

	def get_buffer(self):
		return self.textbuffer

gobject.type_register(TextManager)

gtk.binding_entry_add_signal(TextManager, gtk.keysyms.Return, 0, 'enter-pressed', gobject.TYPE_INT, 0)
gtk.binding_entry_add_signal(TextManager, gtk.keysyms.KP_Enter, 0, 'enter-pressed', gobject.TYPE_INT, 0)

gtk.binding_entry_add_signal(TextManager, gtk.keysyms.Return, gtk.gdk.SHIFT_MASK, 'enter-pressed', gobject.TYPE_INT, gtk.gdk.SHIFT_MASK)
gtk.binding_entry_add_signal(TextManager, gtk.keysyms.KP_Enter, gtk.gdk.SHIFT_MASK, 'enter-pressed', gobject.TYPE_INT, gtk.gdk.SHIFT_MASK)

gtk.binding_entry_add_signal(TextManager, gtk.keysyms.Return, gtk.gdk.CONTROL_MASK, 'enter-pressed', gobject.TYPE_INT, gtk.gdk.CONTROL_MASK)
gtk.binding_entry_add_signal(TextManager, gtk.keysyms.KP_Enter, gtk.gdk.CONTROL_MASK, 'enter-pressed', gobject.TYPE_INT, gtk.gdk.CONTROL_MASK)

gtk.binding_entry_add_signal(TextManager, gtk.keysyms.Return, gtk.gdk.MOD1_MASK, 'enter-pressed', gobject.TYPE_INT, gtk.gdk.MOD1_MASK)
gtk.binding_entry_add_signal(TextManager, gtk.keysyms.KP_Enter, gtk.gdk.MOD1_MASK, 'enter-pressed', gobject.TYPE_INT, gtk.gdk.MOD1_MASK)
